%%% ------------------------------------------------------------------------------------------------
%%% ETCP is available for use under the following license, commonly known as the 3-clause (or
%%% "modified") BSD license:
%%%
%%% Copyright (c) 2017-2018, Pouriya Jahanbakhsh
%%% (pouriya.jahanbakhsh@gmail.com)
%%% All rights reserved.
%%%
%%% Redistribution and use in source and binary forms, with or without modification, are permitted
%%% provided that the following conditions are met:
%%%
%%% 1. Redistributions of source code must retain the above copyright notice, this list of
%%%    conditions and the following disclaimer.
%%%
%%% 2. Redistributions in binary form must reproduce the above copyright notice, this list of
%%%    conditions and the following disclaimer in the documentation and/or other materials provided
%%%    with the distribution.
%%%
%%% 3. Neither the name of the copyright holder nor the names of its contributors may be used to
%%%    endorse or promote products derived from this software without specific prior written
%%%    permission.
%%%
%%% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
%%% IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
%%% FITNESS FOR A  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
%%% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
%%% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
%%% SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
%%% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
%%% OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
%%% POSSIBILITY OF SUCH DAMAGE.
%%% ------------------------------------------------------------------------------------------------
%% @author  Pouriya Jahanbakhsh <pouriya.jahanbakhsh@gmail.com>
%% @version 17.9.10
%% @hidden
%% -------------------------------------------------------------------------------------------------
-module(etcp_acceptor_sup).
-author("pouriya.jahanbakhsh@gmail.com").
%% -------------------------------------------------------------------------------------------------
%% Exports:

%% API:
-export([start_link/5
        ,fetch/1
        ,sleep/1
        ,wakeup/1
        ,fetch_mode/1
        ,start_acceptor/1]).

%% 'director' callback:
-export([init/1
        ,terminate/2]).

%% -------------------------------------------------------------------------------------------------
%% Records & Macros & Includes:

-define(DEF_START_OPTS, []).

-include("internal/etcp_guard.hrl").

%% -------------------------------------------------------------------------------------------------
%% API functions:

-spec
start_link(module(), term(), etcp_types:start_options(), etcp_types:socket(), pid()) ->
    etcp_types:start_return().
start_link(Mod, InitArg, Opts, LSock, Pool) when erlang:is_map(Opts) andalso ?is_proc_ref(Pool) ->
    director:start_link(?MODULE, {Mod, InitArg, Opts, LSock, Pool}, ?DEF_START_OPTS).


-spec
fetch(etcp_types:name()) ->
    {'ok', {reference(), pid()}} | {'error', 'not_found'}.
%% @doc
%%      returns all acceptors.
%% @end
fetch(AccSup) when ?is_proc_ref(AccSup) ->
    case director:get_pids(AccSup) of
        {ok, [Acc]} ->
            {ok, Acc};
        {ok, []} ->
            {error, not_found}
    end.


-spec
sleep(etcp_types:name()) ->
    'ok' | {'error', 'not_found'}.
sleep(AccSup) when ?is_proc_ref(AccSup) ->
    case fetch(AccSup) of
        {ok, {_, Pid}} ->
            etcp_connection:sleep(Pid);
        {error, not_found}=Err ->
            Err
    end.


-spec
wakeup(etcp_types:name()) ->
    'ok' | {'error', 'not_found'}.
wakeup(AccSup) when ?is_proc_ref(AccSup) ->
    case fetch(AccSup) of
        {ok, {_, Pid}} ->
            etcp_connection:wakeup(Pid);
        {error, not_found}=Err ->
            Err
    end.


-spec
fetch_mode(etcp_types:name()) ->
    etcp_types:acceptor_mode() | {'error', 'not_found'}.
fetch_mode(AccSup) when ?is_proc_ref(AccSup) ->
    case fetch(AccSup) of
        {ok, {_, Pid}} ->
            etcp_connection:fetch_mode(Pid);
        {error, not_found}=Err ->
            Err
    end.


-spec
start_acceptor(etcp_types:name()) ->
    'ok' | etcp_types:error().
start_acceptor(AccSup) when ?is_proc_ref(AccSup) ->
    case director:start_child(AccSup, #{start => {etcp_connection, start_link, []}
                                       ,id => erlang:make_ref()
                                       ,append => true}) of
        {ok, _} ->
            ok;
        {error, _}=Err ->
            Err
    end.

%% -------------------------------------------------------------------------------------------------
%% 'director' callback:

%% @hidden
init({Mod, InitArg, Opts, LSock, Pool}) ->
    DefChildSpec = #{start => {etcp_connection, start_link, [Mod, InitArg, Opts, LSock, Pool]}
                    ,plan => fun(_, _, _, St) -> {stop, St} end
                    ,log_validator => fun child_log_validator/4},
    Opts2 = [{log_validator, fun sup_log_validator/4}],
    {ok, undefined, [#{start => {etcp_connection, start_link, []}
                      ,id => erlang:make_ref()
                      ,append => true}], DefChildSpec, Opts2}.


%% @hidden
terminate(_Rsn, _State) ->
    ok.

%% -------------------------------------------------------------------------------------------------
%% Internal functions:

%%get_mode_fix_return([{_, Mode}|Modes]=Modes2) ->
%%    get_mode_fix_return(Modes, Modes2, Mode).
%%
%%
%%get_mode_fix_return([{_, Mode}|Modes], Modes2, Mode) ->
%%    get_mode_fix_return(Modes, Modes2, Mode);
%%get_mode_fix_return([], _Modes, Mode) ->
%%    Mode;
%%get_mode_fix_return([_|_Modes], Modes2, _Mode) ->
%%    Modes2.


child_log_validator(_Name, info, start, _State) ->
    none;
child_log_validator(_Name, error, normal, _State) ->
    none;
child_log_validator(_Name, error, _, _State) ->
    long.


sup_log_validator(_Name, error, _, _State) ->
    none;
sup_log_validator(_Name, warning, _, _State) ->
    long.