%%% ------------------------------------------------------------------------------------------------
%%% ETCP is available for use under the following license, commonly known as the 3-clause (or
%%% "modified") BSD license:
%%%
%%% Copyright (c) 2017-2018, Pouriya Jahanbakhsh
%%% (pouriya.jahanbakhsh@gmail.com)
%%% All rights reserved.
%%%
%%% Redistribution and use in source and binary forms, with or without modification, are permitted
%%% provided that the following conditions are met:
%%%
%%% 1. Redistributions of source code must retain the above copyright notice, this list of
%%%    conditions and the following disclaimer.
%%%
%%% 2. Redistributions in binary form must reproduce the above copyright notice, this list of
%%%    conditions and the following disclaimer in the documentation and/or other materials provided
%%%    with the distribution.
%%%
%%% 3. Neither the name of the copyright holder nor the names of its contributors may be used to
%%%    endorse or promote products derived from this software without specific prior written
%%%    permission.
%%%
%%% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
%%% IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
%%% FITNESS FOR A  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
%%% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
%%% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
%%% SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
%%% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
%%% OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
%%% POSSIBILITY OF SUCH DAMAGE.
%%% ------------------------------------------------------------------------------------------------
%% @author  Pouriya Jahanbakhsh <pouriya.jahanbakhsh@gmail.com>
%% @version 17.9.10
%% -------------------------------------------------------------------------------------------------
-module(etcp_connection).
-author("pouriya.jahanbakhsh@gmail.com").
%% -------------------------------------------------------------------------------------------------
%% Exports:

%% API:
-export([start_link/4
        ,start_link/5
        ,start_link/6
        ,send_sync/2
        ,send_sync/3
        ,send_async/2
        ,stop/1
        ,stop/2
        ,stop/3

        ,sleep/1
        ,wakeup/1
        ,fetch_mode/1
        ,change_mode/1]).

%% 'gen' callback:
-export([init_it/6]).

%% 'sys' callbacks:
-export([system_terminate/4
        ,system_replace_state/2
        ,system_get_state/1
        ,system_continue/3
        ,system_code_change/4]).

%% 'proc_lib' callback:
-export([wake_hibernate/3]).

%% -------------------------------------------------------------------------------------------------
%% Records & Macros & Includes:

-define(STATE, etcp_connection_state).
-record(?STATE, {name, data, active, module, metadata, hibernate = false}).

-define(A_STATE, etcp_acceptor_state).
-record(?A_STATE, {module, init_argument, name, options, metadata, mode, connection_debug, pool}).

%% Dependencies:
%%  #etcp_metadata{}
-include("internal/etcp_metadata.hrl").
-define(ETCP_MD, etcp_metadata).

-define(DEF_TERMINATE_TIMEOUT, 5000).
-define(GEN_CALL_TAG, '$gen_call').
-define(GEN_CAST_TAG, '$gen_cast').
-define(GEN_EVENT_TAG, '$gen_event').
-define(GEN_ETCP_TAG, '$gen_etcp').
-define(DEF_LENGHT, 0).
-define(DEF_TIMEOUT, infinity).
-define(DEF_SRTIMEOUT, infinity).
-define(DEF_START_OPTS, #{}).
-define(DEF_PROC_START_OPTS, []).
-define(DEF_SOCKET_OPTIONS, []).
-define(DEF_SSL_FLAG, false).
-define(DEF_CONNECT_TIMEOUT, 3000).
-define(DEF_DEBUG, []).
-define(DEF_TRANSPORT_MODULE, 'etcp_transporter_tcp').
-define(DEF_TERMINATE_LOG, 'true').

-define(SLEEP, 'sleep').
-define(ACCEPT, 'accept').
-define(DEF_MODE, ?ACCEPT).
-define(CHANGE_MODE_TAG, 'change_mode').
-define(DEF_RECEIVE_TIMEOUT, 300).
-define(FETCH_MODE_TAG, 'fetch_mode').

-include("internal/etcp_guard.hrl").

%% -------------------------------------------------------------------------------------------------
%% API:

-spec
start_link(module(), term(), etcp_types:host(), etcp_types:port_number()) ->
    etcp_types:start_return().
%% @doc
%%      starts and links a socket connection handler process.
%% @end
start_link(Mod, InitArg, Host, Port) when erlang:is_atom(Mod) andalso
                                          erlang:is_integer(Port) ->
    gen:start(?MODULE
             ,link
             ,Mod
             ,{client, InitArg, Host, Port, ?DEF_START_OPTS}
             ,?DEF_PROC_START_OPTS).


-spec
start_link(module() | etcp_types:register_name()
          ,term() | module()
          ,etcp_types:host()| term() | etcp_types:start_options()
          ,etcp_types:port_number() | etcp_types:host()
          ,etcp_types:options() | etcp_types:port_number()) ->
    etcp_types:start_return().
start_link(Mod, InitArg, Host, Port, Opts) when erlang:is_atom(Mod) andalso
                                                erlang:is_integer(Port) andalso
                                                erlang:is_map(Opts) ->
    gen:start(?MODULE, link, Mod, {client, InitArg, Host, Port, Opts}, ?DEF_PROC_START_OPTS);
start_link(Name, Mod, InitArg, Host, Port) when erlang:is_tuple(Name) andalso
                                                erlang:is_atom(Mod) andalso
                                                erlang:is_integer(Port) ->
    gen:start(?MODULE
             ,link
             ,Name
             ,Mod
             ,{client, InitArg, Host, Port, ?DEF_START_OPTS}
             ,?DEF_PROC_START_OPTS);
%% Will called by calling etcp_connection_sup:add/3
start_link(Mod, InitArg, Opts, LSock, Pool) when erlang:is_atom(Mod) andalso
                                                 erlang:is_map(Opts) andalso
                                                 erlang:is_pid(Pool) ->
    gen:start(?MODULE, link, Mod, {server, InitArg, Opts, LSock, Pool}, ?DEF_PROC_START_OPTS).


-spec
start_link(etcp_types:register_name()
          ,module()
          ,term()
          , etcp_types:host()
          , etcp_types:port_number()
          , etcp_types:options()) ->
    etcp_types:start_return().
start_link(Name, Mod, InitArg, Host, Port, Opts) when erlang:is_tuple(Name) andalso
                                                      erlang:is_atom(Mod) andalso
                                                      erlang:is_integer(Port) andalso
                                                      erlang:is_map(Opts) ->
    gen:start(?MODULE, link, Name, Mod, {client, InitArg, Host, Port, Opts}, ?DEF_PROC_START_OPTS).


-spec
send_sync(etcp_types:name(), etcp_types:packet()) ->
    'ok' | etcp_types:error().
send_sync(Con, Packet) when ?is_proc_ref(Con) ->
    case catch gen:call(Con, ?GEN_ETCP_TAG, {send, Packet}) of
        {ok, Res} ->
            Res;
        {'EXIT', Reason} ->
            erlang:exit({Reason, [{module, ?MODULE}
                                 ,{function, send_sync}
                                 ,{arguments, [Con, Packet]}]})
    end.


-spec
send_sync(etcp_types:name(), etcp_types:packet(), timeout()) ->
    'ok' | etcp_types:error().
send_sync(Con, Packet, Timeout) when ?is_proc_ref(Con) andalso ?is_timeout(Timeout) ->
    case catch gen:call(Con, ?GEN_ETCP_TAG, {send, Packet}, Timeout) of
        {ok, Res} ->
            Res;
        {'EXIT', Reason} ->
            erlang:exit({Reason, [{module, ?MODULE}
                                 ,{function, send_sync}
                                 ,{arguments, [Con, Packet, Timeout]}]})
    end.


-spec
send_async(etcp_types:name(), etcp_types:packet()) ->
    'ok'.
send_async(Con, Packet) when ?is_proc_ref(Con) ->
    catch Con ! {?GEN_ETCP_TAG, {send, Packet}},
    ok.


-spec
stop(etcp_types:name()) ->
    ok.
stop(Con) when ?is_proc_ref(Con) ->
    proc_lib:stop(Con, normal, ?DEF_TERMINATE_TIMEOUT).


-spec
stop(etcp_types:name(), etcp_type:reason()) ->
    ok.
stop(Con, Reason) when ?is_proc_ref(Con) ->
    proc_lib:stop(Con, Reason, ?DEF_TERMINATE_TIMEOUT).


-spec
stop(etcp_types:name(), etcp_type:reason(), timeout()) ->
  ok.
stop(Con, Reason, Timeout) when ?is_proc_ref(Con) andalso ?is_timeout(Timeout) ->
  proc_lib:stop(Con, Reason, Timeout).


-spec
sleep(etcp_types:name()) ->
    'ok'.
sleep(Acc) when ?is_proc_ref(Acc) ->
    gen_server:call(Acc, ?SLEEP).


-spec
wakeup(etcp_types:name()) ->
    'ok'.
wakeup(Acc) when ?is_proc_ref(Acc) ->
    gen_server:call(Acc, ?ACCEPT).


-spec
fetch_mode(etcp_types:name()) ->
    etcp_types:acceptor_mode().
fetch_mode(Acc) when ?is_proc_ref(Acc) ->
    gen_server:call(Acc, ?FETCH_MODE_TAG).


-spec
change_mode(etcp_types:name()) ->
    etcp_types:acceptor_mode().
change_mode(Acc) when ?is_proc_ref(Acc) ->
    gen_server:call(Acc, ?CHANGE_MODE_TAG).

%% -------------------------------------------------------------------------------------------------
%% 'gen' callback:

%% @hidden
init_it(Starter, Parent, Name, Mod, {client, InitArg, Host, Port, Opts}, _Opts) ->
    TrMod = etcp_utils:get_value(transporter, Opts, ?DEF_TRANSPORT_MODULE, fun erlang:is_atom/1),
    case etcp_transporter:connect(TrMod, Host, Port, Opts) of
        {ok, Sock} ->
            TrMod = etcp_utils:get_value(transporter
                                        ,Opts
                                        ,?DEF_TRANSPORT_MODULE
                                        ,fun erlang:is_atom/1),
            DbgOpts = etcp_utils:get_value(connection_debug
                                          ,Opts
                                          ,?DEF_DEBUG
                                          ,fun erlang:is_list/1),
            Dbg = etcp_utils:debug_options(?MODULE, Name, DbgOpts),
            case fix_activity(TrMod, Sock, Opts) of
                {ok, Activity} ->
                    TMD = etcp_metadata:wrap(Sock
                                            ,?DEF_TIMEOUT
                                            ,?DEF_SRTIMEOUT
                                            ,?DEF_LENGHT
                                            ,TrMod
                                            ,Opts
                                            ,undefined
                                            ,undefined
                                            ,{Host, Port}),
                    State = #?STATE{name = Name
                                   ,data = undefined
                                   ,module = Mod
                                   ,active = Activity
                                   ,metadata = TMD
                                   ,hibernate = false},
                    case run_callback2(Dbg, State, Mod, connection_init, [InitArg, TMD]) of
                        {ok, Dbg2, State2} ->
                            proc_lib:init_ack(Starter, {ok, erlang:self()}),
                            connection_loop(Parent, Dbg2, State2);

                        {close, _Dbg2, _State2} ->
                            _ = etcp_transporter:close(TrMod, Sock, Opts),
                            proc_lib:init_ack(Starter, {error, normal}),
                            erlang:exit(normal);

                        {stop, _Dbg2, _State2, Reason} ->
                            _ = etcp_transporter:close(TrMod, Sock, Opts),
                            proc_lib:init_ack(Starter, {error, Reason}),
                            erlang:exit(Reason);

                        {error, {callback_return, [{returned_value, ignore}|_]}} ->
                            _ = etcp_transporter:close(TrMod, Sock, Opts),
                            proc_lib:init_ack(Starter, ignore),
                            erlang:exit(normal);

                        {error, Reason}=Error ->
                            _ = etcp_transporter:close(TrMod, Sock, Opts),
                            proc_lib:init_ack(Starter, Error),
                            erlang:exit(Reason)
                    end;
                {error, Reason}=Error ->
                    _ = etcp_transporter:close(TrMod, Sock, Opts),
                    proc_lib:init_ack(Starter, Error),
                    erlang:exit(Reason)
            end;
        {error, Reason}=Error ->
            proc_lib:init_ack(Starter, Error),
            erlang:exit(Reason)
    end;
init_it(Starter, Parent, Name, Mod, {server, InitArg, Opts, LSock, Pool}, _Opts) ->
    DbgOpts = etcp_utils:get_value(acceptor_debug, Opts, ?DEF_DEBUG, fun erlang:is_list/1),
    Dbg = etcp_utils:debug_options(?MODULE, Name, DbgOpts),
    TrMod = etcp_utils:get_value(transporter, Opts, ?DEF_TRANSPORT_MODULE, fun erlang:is_atom/1),
    Mode = etcp_utils:get_value(acceptor_mode, Opts, ?DEF_MODE, fun filter_mode/1),
    ConDbgOpts = etcp_utils:get_value(connection_debug, Opts, ?DEF_DEBUG, fun erlang:is_list/1),
    proc_lib:init_ack(Starter, {ok, erlang:self()}),
    Dbg2 = acceptor_debug(Name, Dbg, {start, TrMod, LSock, Pool, Mode}),
    Metadata = etcp_metadata:wrap(LSock
                                 ,?DEF_TIMEOUT
                                 ,?DEF_SRTIMEOUT
                                 ,?DEF_LENGHT
                                 ,TrMod
                                 ,Opts
                                 ,undefined
                                 ,undefined
                                 ,{undefined, undefined}),
    State = #?A_STATE{module = Mod
                     ,init_argument = InitArg
                     ,metadata = Metadata
                     ,name = Name
                     ,options = Opts
                     ,mode = Mode
                     ,connection_debug = ConDbgOpts
                     ,pool = Pool},
    acceptor_loop(Parent, Dbg2, State).

%% -------------------------------------------------------------------------------------------------
%% 'sys' callbacks:

%% @hidden
system_code_change(#?STATE{module = Mod, data = Data}=State, _Module, OldVsn, Extra) ->
    case Mod:code_change(OldVsn, Data, Extra) of
        {ok, Data2} ->
            {ok, State#?STATE{data = Data2}};
        Other ->
            Other
    end.


%% @hidden
system_continue(Parent, Dbg, State) when erlang:is_record(State, ?STATE) ->
    connection_loop(Parent, Dbg, State);
system_continue(Parent, Dbg, State) when erlang:is_record(State, ?A_STATE) ->
    acceptor_loop(Parent, Dbg, State).


%% @hidden
system_get_state(#?STATE{data=Data}) ->
    {ok, Data}.


%% @hidden
system_replace_state(StateFun, #?STATE{data=Data}=State) ->
    Data2 = StateFun(Data),
    {ok, Data2, State#?STATE{data=Data2}}.


%% @hidden
system_terminate(Reason, _Parent, Dbg, State) ->
    terminate(Dbg, State, Reason).

%% -------------------------------------------------------------------------------------------------
%% Internal functions:

connection_loop(Parent
               ,Dbg
               ,#?STATE{name = Name
                       ,active = false
                       ,metadata = #?ETCP_MD{socket =Sock
                                            ,length = Len
                                            ,srtimeout = SRTimeout
                                            ,transporter = TrMod
                                            ,options = Opts}}=State) when Len >= 0 ->
    case socket_receive(TrMod, Sock, Name, Dbg, Len, SRTimeout, Opts) of
        {ok, Dbg2, Packet} ->
            {Dbg3, State2} = run_callback(Dbg2, State, handle_packet, {Packet}),
            connection_receive(Parent, Dbg3, State2);
        {error, {socket_recv, [{reason, timeout}|_ErrorParams]}} ->
            {Dbg2, State2} = run_callback(Dbg, State, srtimeout, {}),
            connection_receive(Parent, Dbg2, State2);
        {error, {socket_recv, [{reason, closed}|_ErrorParams]}} ->
            {Dbg2, State2} = run_callback(Dbg,State, handle_disconnect, {}),
            terminate(Dbg2, State2, normal);
        {error, Reason} ->
            terminate(Dbg, State, Reason)
    end;
connection_loop(Parent, Dbg, State) ->
    connection_receive(Parent, Dbg, State).


acceptor_loop(Parent, Dbg, #?A_STATE{mode = ?ACCEPT
                                    ,metadata = #etcp_metadata{transporter = TrMod
                                                              ,socket = LSock
                                                              ,options = Opts}=MD
                                    ,name = Name
                                    ,module = Mod
                                    ,init_argument = InitArg
                                    ,connection_debug = ConDbgOpts
                                    ,pool = Pool}=State) ->
    case etcp_transporter:accept(TrMod, LSock, Opts) of
        {ok, Sock} ->
            Dbg2 = acceptor_debug(Name, Dbg, {accept, Sock}),
            try
                {director:delete_running_child(Parent, erlang:self())
                ,director:become_supervisor(Pool
                                           ,#{id => erlang:make_ref()
                                             ,start => {?MODULE
                                                       ,start_link
                                                       ,[Mod, InitArg, Sock, Opts]}
                                             ,append => true}
                                           ,erlang:self())
                ,etcp_acceptor_sup:start_acceptor(Parent)} of
                {ok, ok, ok} ->
                    _Dbg3 = acceptor_debug(Name, Dbg2, {start_connection, Sock, Parent}),
                    receive
                        {change_parent, _} ->
                            ok
                    end,
                    case etcp_transporter:activity(TrMod, Sock, Opts) of
                        {ok, Activity} ->
                            ConState = #?STATE{name = Name
                                              ,data = undefined
                                              ,module = Mod
                                              ,active = Activity
                                              ,metadata = MD#etcp_metadata{socket = Sock}
                                              ,hibernate = false},
                            ConDbg = etcp_utils:debug_options(?MODULE, Name, ConDbgOpts),
                            case run_callback2(ConDbg
                                              ,ConState
                                              ,Mod
                                              ,connection_init
                                              ,[InitArg, MD]) of
                                {ok, ConDbg2, ConState2} ->
                                    connection_loop(Pool, ConDbg2, ConState2);
                                {close, _ConDbg, _ConState2} ->
                                    _ = etcp_transporter:close(TrMod, Sock, Opts),
                                    erlang:exit(normal);
                                {stop, _ConDbg, _ConState2, Reason} ->
                                    _ = etcp_transporter:close(TrMod, Sock, Opts),
                                    erlang:exit(Reason);
                                {error, {callback_return, [{returned_value, ignore}|_]}} ->
                                    _ = etcp_transporter:close(TrMod, Sock, Opts),
                                    erlang:exit(normal);
                                {error, Reason} ->
                                    _ = etcp_transporter:close(TrMod, Sock, Opts),
                                    erlang:exit(Reason)
                            end;
                        {error, Reason} ->
                            _ = etcp_transporter:close(TrMod, Sock, Opts),
                            erlang:exit(Reason)
                    end;
                {Result1, Result2, Result3} ->
                    terminate(Dbg2, State, {change_parent, [{delete_running_child, Result1}
                                                           ,{become_supervisor, Result2}
                                                           ,{start_child, Result3}]})
            catch
                _:Rsn ->
                    terminate(Dbg2, State, {change_parent, [{reason, Rsn}
                                                           ,{stacktrace, erlang:get_stacktrace()}
                                                           ,{parent, Parent}]})
            end;
        {error, {socket_accept, [{reason, timeout}|_ErrorParams]}} ->
            acceptor_receive(Parent, Dbg, State);
        {error, {socket_accept, [{reason, {handshake, timeout}}|_ErrorParams]}} ->
            acceptor_receive(Parent, Dbg, State);
        {error, {socket_accept, [{reason, emfile}|_]}} ->
            error_logger:warning_msg("ETCP acceptor ~p: too many open connections", [Name]),
            acceptor_receive(Parent
                            ,acceptor_debug(Name, Dbg, {?CHANGE_MODE_TAG, ?SLEEP})
                            ,State#?A_STATE{module = ?SLEEP});
        {error, Reason} ->
            terminate(Dbg, State, Reason)
    end;
acceptor_loop(Parent, Dbg, State) ->
    acceptor_receive(Parent, Dbg, State).


connection_receive(Parent, Dbg, #?STATE{hibernate = false
                                       ,metadata = #?ETCP_MD{timeout = Timeout}=TMD}=State) ->
    receive
        Msg ->
            connection_process_message(Parent
                                      ,Dbg
                                      ,State#?STATE{metadata = TMD#?ETCP_MD{last_message = Msg}}
                                      ,Msg)
    after Timeout ->
        {Dbg2, State2} = run_callback(Dbg, State, timeout, {}),
        connection_loop(Parent, Dbg2, State2)
    end;
connection_receive(Parent, Dbg, State) ->
    proc_lib:hibernate(?MODULE, wake_hibernate, [Parent, Dbg, State#?STATE{hibernate = false}]).


wake_hibernate(Parent, Dbg, #?STATE{hibernate = false
                                   ,metadata = #?ETCP_MD{timeout = Timeout}=TMD}=State) ->
    receive
        Msg ->
            connection_process_message(Parent
                                      ,Dbg
                                      ,State#?STATE{metadata = TMD#?ETCP_MD{last_message = Msg}}
                                      ,Msg)
    after Timeout ->
        {Dbg2, State2} = run_callback(Dbg, State, timeout, {}),
        connection_loop(Parent, Dbg2, State2)
    end.


connection_process_message(Parent, Dbg, #?STATE{name=Name}=State, {?GEN_ETCP_TAG, From, Req}=Msg) ->
    {Dbg2, State2} = connection_process_sync_request(connection_debug(Name, Dbg, Msg), State, From, Req),
    connection_loop(Parent, Dbg2, State2);
connection_process_message(Parent, Dbg, #?STATE{name=Name}=State, {?GEN_ETCP_TAG, Req}=Msg) ->
    {Dbg2, State2} = connection_process_async_request(connection_debug(Name, Dbg, Msg), State, Req),
    connection_loop(Parent, Dbg2, State2);
connection_process_message(Parent, Dbg, #?STATE{name= Name}=State, {?GEN_CALL_TAG, From, Req}=Msg) ->
    {Dbg2, State2} = run_callback(connection_debug(Name, Dbg, Msg), State, handle_call, {Req, From}),
    connection_loop(Parent, Dbg2, State2);
connection_process_message(Parent, Dbg, #?STATE{name= Name}=State, {?GEN_CAST_TAG, Msg2}=Msg) ->
    {Dbg2, State2} = run_callback(connection_debug(Name, Dbg, Msg), State, handle_cast, {Msg2}),
    connection_loop(Parent, Dbg2, State2);
connection_process_message(Parent, Dbg, #?STATE{name= Name}=State, {?GEN_EVENT_TAG, Event}=Msg) ->
    {Dbg2, State2} = run_callback(connection_debug(Name, Dbg, Msg), State, handle_event, {Event}),
    connection_loop(Parent, Dbg2, State2);
connection_process_message(Parent, Dbg, #?STATE{name = Name}=State, {system, From, Msg}=SysMsg) ->
    sys:handle_system_msg(Msg, From, Parent, ?MODULE, connection_debug(Name, Dbg, SysMsg), State);
connection_process_message(Parent, Dbg, State, {'EXIT', Parent, Reason}) ->
    terminate(Dbg, State, Reason);
connection_process_message(Parent
                          ,Dbg
                          ,#?STATE{name = Name, metadata = #?ETCP_MD{transporter = TrMod
                                                                    ,options = Opts
                                                                    ,socket = Sock}}=State
                          ,Msg) ->
    case etcp_transporter:check_message(TrMod, Msg, Sock, Opts) of
        {ok, Packet} ->
            {Dbg2, State2} = run_callback(connection_debug(Name, Dbg, {socket_in, Packet})
                                         ,State
                                         ,handle_packet
                                         ,{Packet}),
            connection_loop(Parent, Dbg2, State2);
        unknown ->
            {Dbg2, State2} = run_callback(connection_debug(Name, Dbg, {in, Msg}), State, handle_info, {Msg}),
            connection_loop(Parent, Dbg2, State2);
        {error, {socket_check_message, [{reason, closed}|_]}} ->
            {Dbg2, State2} = run_callback(Dbg, State, handle_disconnect, {}),
            terminate(Dbg2, State2, normal);
        {error, Reason} ->
            terminate(Dbg, State, Reason)
    end.


connection_process_sync_request(Dbg
                               ,#?STATE{name = Name
                                       ,metadata = #?ETCP_MD{socket = Sock
                                                            ,transporter = TrMod
                                                            ,options = Opts}}=State
                               ,From
                               ,{send, Packet}) ->
    case socket_send(TrMod, Sock, Name,  Dbg, Packet, Opts) of
        {ok, Dbg2} ->
            {connection_reply(Name, Dbg2, From, ok), State};
        {error, Reason}=Error ->
            terminate(connection_reply(Name, Dbg, From, Error), State, Reason)
    end;
connection_process_sync_request(Dbg, #?STATE{name = Name}=State, From, Other) ->
    {connection_reply(Name, Dbg, From, {error, {unknown_request, [{request, Other}]}}), State}.


connection_process_async_request(Dbg
                                ,#?STATE{name = Name
                                        ,metadata = #?ETCP_MD{socket = Sock
                                                             ,transporter = TrMod
                                                             ,options = Opts}}=State
                                ,{send, Packet}) ->
    case socket_send(TrMod, Sock, Name,  Dbg, Packet, Opts) of
        {ok, Dbg2} ->
            {Dbg2, State};
        {error, Reason} ->
            terminate(Dbg, State, Reason)
    end;
connection_process_async_request(Dbg, State, Other) ->
    terminate(Dbg, State, {unknown_request, [{request, Other}]}).


acceptor_receive(Parent, Dbg, #?A_STATE{mode = Mode}=State) ->
    Timeout =
        if
            Mode == ?ACCEPT ->
                0;
            true ->
                ?DEF_RECEIVE_TIMEOUT
        end,
    receive
        Msg ->
            acceptor_process_message(Parent, Dbg, State, Msg)
    after Timeout ->
        acceptor_loop(Parent, Dbg, State)
    end.


acceptor_process_message(Parent, Dbg, #?A_STATE{name=Name}=State, {?GEN_CALL_TAG, From, Msg}) ->
    {Dbg2, State2} = acceptor_process_request(acceptor_debug(Name, Dbg, {?GEN_CALL_TAG, From, Msg})
                                             ,State
                                             ,From
                                             ,Msg),
    acceptor_loop(Parent, Dbg2, State2);
acceptor_process_message(Parent, Dbg, State, {system, From, Msg}) ->
    sys:handle_system_msg(Msg, From, Parent, ?MODULE, Dbg, State);
acceptor_process_message(Parent, Dbg, State, {'EXIT', Parent, Reason}) ->
    terminate(Dbg, State, Reason);
acceptor_process_message(Parent, Dbg, #?A_STATE{name= Name}=State, Msg) ->
    acceptor_loop(Parent, acceptor_debug(Name, Dbg, {in, Msg}), State).


acceptor_process_request(Dbg, #?A_STATE{name = Name}=State, From, ?SLEEP) ->
    {acceptor_reply(Name, acceptor_debug(Name, Dbg, {?CHANGE_MODE_TAG, ?SLEEP}), From, ok)
    ,State#?A_STATE{mode = ?SLEEP}};
acceptor_process_request(Dbg, #?A_STATE{name = Name}=State, From, ?ACCEPT) ->
    {acceptor_reply(Name, acceptor_debug(Name, Dbg, {?CHANGE_MODE_TAG, ?ACCEPT}), From, ok)
    ,State#?A_STATE{mode = ?ACCEPT}};
acceptor_process_request(Dbg, #?A_STATE{name = Name, mode = Mode}=State, From, ?FETCH_MODE_TAG) ->
    {acceptor_reply(Name, acceptor_debug(Name, Dbg, ?FETCH_MODE_TAG), From, Mode), State};
acceptor_process_request(Dbg, #?A_STATE{name = Name, mode = Mode}=State, From, ?CHANGE_MODE_TAG) ->
    Mode2 =
        case Mode of
            ?SLEEP ->
                ?ACCEPT;
            ?ACCEPT ->
                ?SLEEP
        end,
    {acceptor_reply(Name
                   ,acceptor_debug(Name, Dbg, {?CHANGE_MODE_TAG, Mode2})
                   ,From
                   ,Mode2)
    ,State#?A_STATE{mode = Mode2}};
acceptor_process_request(Dbg, #?A_STATE{name = Name}=State, From, Other) ->
    {acceptor_reply(Name, Dbg, From, {error, {unknown_call, [{call, Other}]}}), State}.


connection_reply(Name, Dbg, {Pid, Tag}=Client, Msg) ->
    catch Pid ! {Tag, Msg},
    connection_debug(Name, Dbg, {out, Client, Msg});
connection_reply(Name, Dbg, Pid, Msg) ->
    connection_debug(Name, Dbg, {out, Pid, Msg}).

acceptor_reply(Name, Dbg, {Pid, Tag}, Msg) ->
    catch Pid ! {Tag, Msg},
    acceptor_debug(Name, Dbg, {out, Pid, Msg});
acceptor_reply(Name, Dbg, Pid, Msg) ->
    acceptor_debug(Name, Dbg, {out, Pid, Msg}).


run_callback(Dbg, #?STATE{module = Mod, data= Data, metadata = TMD}=State, Callback, Args) ->
    Args2 =
        case Args of
            {} ->
                [Data, TMD];
            {Arg} ->
                [Arg, Data, TMD];
            {Arg1, Arg2} ->
                [Arg1, Arg2, Data, TMD]
        end,
    case run_callback2(Dbg, State, Mod, Callback, Args2) of
        {ok, Dbg2, State2} ->
            {Dbg2, State2};
        {close, Dbg2, State2} ->
            terminate(Dbg2, State2, normal);
        {stop, Dbg2, State2, Reason} ->
            terminate(Dbg2, State2, Reason);
        {error, Reason} ->
            terminate(Dbg, State, Reason)
    end.


run_callback2(Dbg, #?STATE{metadata = TMD}=State, Mod, Func, Args) ->
    Ret =
        try erlang:apply(Mod, Func, Args) of
            ok ->
                {ok, Dbg, State#?STATE{metadata = TMD#?ETCP_MD{last_callback = Func}}};
            {ok, Opts} ->
                get_options(Dbg, State, Opts);
            close ->
                {close, Dbg, State#?STATE{metadata = TMD#?ETCP_MD{last_callback = Func}}};
            {close, Opts} ->
                case get_options(Dbg, State, Opts) of
                    {ok, Dbg2, State2}  ->
                        {close, Dbg2, State2#?STATE{metadata = TMD#?ETCP_MD{last_callback = Func}}};
                    {error, _}=Err ->
                        Err
                end;
            {stop, Reason} ->
                {stop, Dbg, State#?STATE{metadata = TMD#?ETCP_MD{last_callback = Func}},Reason};
            {stop, Reason, Opts} ->
                case get_options(Dbg, State, Opts) of
                    {ok, Dbg2, State2}  ->
                        {stop
                        ,Dbg2
                        ,State2#?STATE{metadata = TMD#?ETCP_MD{last_callback = Func}}
                        ,Reason};
                    {error, _}=Err ->
                        Err
                end;
            Other ->
                {error, {callback_return, [{returned_value, Other}]}}
        catch
            _:Rsn ->
                {error, {callback_crash, [{reason, Rsn}, {stacktrace, erlang:get_stacktrace()}]}}
        end,
    case Ret of
        {error, {Rsn2, ErrParams}} ->
            {error, {Rsn2, ErrParams ++ [{module, Mod}, {function, Func}, {arguments, Args}]}};
        _ ->
            Ret
    end.


get_options(Dbg, State, [H|T]) when not erlang:is_list(T) ->
    %% Fix improper list:
    get_options(Dbg, State, [H,T]);
get_options(Dbg, State, [{state, Data} | Opts]) ->
    get_options(Dbg, State#?STATE{data = Data}, Opts);
get_options(Dbg
           ,#?STATE{name = Name
                   ,metadata = #?ETCP_MD{socket = Sock, transporter = TrMod, options = Opts}}=State
           ,[{packet, Pkt} | Opts2]) ->
    case socket_send(TrMod, Sock, Name, Dbg, Pkt, Opts) of
        {ok, Dbg2} ->
            get_options(Dbg2, State, Opts2);
        {error, _Reason}=Error ->
            Error
    end;
get_options(Dbg, #?STATE{metadata = TMD}=State, [{timeout, infinity} | Opts]) ->
    get_options(Dbg, State#?STATE{metadata = TMD#?ETCP_MD{timeout = infinity}}, Opts);
get_options(Dbg
           ,#?STATE{metadata = TMD}=State
           ,[{timeout, Timeout} | Opts]) when erlang:is_integer(Timeout) ->
    if
        Timeout >= 0 ->
            get_options(Dbg, State#?STATE{metadata = TMD#?ETCP_MD{timeout = Timeout}}, Opts);
        true ->
            {error, {option_value, [{timeout, Timeout}]}}
    end;
get_options(Dbg, #?STATE{metadata = TMD}=State, [{srtimeout, infinity} | Opts]) ->
    get_options(Dbg, State#?STATE{metadata = TMD#?ETCP_MD{srtimeout = infinity}}, Opts);
get_options(Dbg
           ,#?STATE{metadata = TMD}=State
           ,[{srtimeout, SRTimeout} | Opts]) when erlang:is_integer(SRTimeout) ->
    if
        SRTimeout >= 0 ->
            get_options(Dbg, State#?STATE{metadata = TMD#?ETCP_MD{srtimeout = SRTimeout}}, Opts);
        true ->
            {error, {option_value, [{srtimeout, SRTimeout}]}}
    end;
get_options(Dbg
           ,#?STATE{metadata = TMD}=State
           ,[{length, Len} |Opts]) when erlang:is_integer(Len) ->
    get_options(Dbg, State#?STATE{metadata = TMD#?ETCP_MD{length = Len}}, Opts);
get_options(Dbg, #?STATE{name = Name}=State, [{reply, From, Msg} | Opts]) ->
    get_options(connection_reply(Name, Dbg, From, Msg), State, Opts);
get_options(Dbg, State, [hibernate|Opts]) ->
    get_options(Dbg, State#?STATE{hibernate = true}, Opts);
get_options(Dbg, State, []) ->
    {ok, Dbg, State};
get_options(Dbg
           ,#?STATE{metadata = TMD}=State
           ,[{transporter, TrMod} | Opts]) when erlang:is_atom(TrMod) ->
    get_options(Dbg, State#?STATE{metadata = TMD#?ETCP_MD{transporter = TrMod}}, Opts);
get_options(Dbg, #?STATE{metadata = TMD}=State, [{socket, Sock} | Opts]) ->
    get_options(Dbg, State#?STATE{metadata = TMD#?ETCP_MD{socket = Sock}}, Opts);
get_options(Dbg
           ,#?STATE{metadata = #?ETCP_MD{transporter = TrMod, socket = Sock, options = Opts}=TMD}=State
           ,[{set_options, SockOpts} | Opts2]) when erlang:is_list(SockOpts) ->
    case etcp_utils:filter_socket_options(SockOpts) of
        ok ->
            case etcp_transporter:set_options(TrMod, Sock, SockOpts, Opts) of
                ok ->
                    case etcp_transporter:activity(TrMod, Sock, Opts) of
                        {ok, Bool} ->
                            TMD2 = TMD#?ETCP_MD{options = Opts#{socket_options => SockOpts}},
                            get_options(Dbg
                                       ,State#?STATE{active = Bool
                                                    ,metadata = TMD2}
                                       ,Opts2);
                        {error, _}=Err ->
                            Err
                    end;
                {error, _}=Err ->
                    Err
            end;
        {error, _}=Err ->
            Err
    end;
get_options(Dbg
           ,#?STATE{metadata = #?ETCP_MD{socket = Sock, transporter = TrMod, options = Opts}=TMD}=State
           ,[fetch_peername | Opts2])  ->
    case etcp_transporter:peername(TrMod, Sock, Opts) of
        {ok, Addr} ->
            get_options(Dbg
                       ,State#?STATE{metadata = TMD#?ETCP_MD{address = Addr}}
                       ,Opts2);
        {error, _}=Err ->
            Err
    end;
%% Catch clauses:
get_options(_Dbg, _State, [{timeout, Timeout} | _Opts]) ->
    {error, {option_value, [{timeout, Timeout}]}};
get_options(_Dbg, _State, [{srtimeout, SRTimeout} | _Opts]) ->
    {error, {option_value, [{srtimeout, SRTimeout}]}};
get_options(_Dbg, _State, [{length, Len} | _Opts]) ->
    {error, {option_value, [{length, Len}]}};
get_options(_Dbg, _State, [{transporter, TrMod}|_Opts]) ->
    {error, {option_value, [{transporter, TrMod}]}};
get_options(_Dbg, _State, [{set_options, Opts}|_Opts]) ->
    {error, {option_value, [{set_options, Opts}]}};
get_options(_Dbg, _State, [{Key, Val}|_Opts]) ->
    {error, {option_value, [{key, Key}, {value, Val}]}};
get_options(_Dbg, _State, [Opt|_Opts]) ->
    {error, {option_value, [{option, Opt}]}};
get_options(_Dbg, _State ,Opts) ->
    {error, {options_value, [{options, Opts}]}}.


socket_receive(TrMod, Sock, Name, Dbg, Len, SRTimeout, Opts) ->
    case etcp_transporter:recv(TrMod, Sock, Len, SRTimeout, Opts) of
        {ok, Packet} ->
            {ok, connection_debug(Name, Dbg, {socket_in, Packet}), Packet};
        {error, _Reason}=Error ->
            Error
    end.


socket_send(TrMod, Sock, Name,  Dbg, Packet, Opts) ->
    case etcp_transporter:send(TrMod, Sock, Packet, Opts) of
        ok ->
            {ok, connection_debug(Name, Dbg, {socket_out, Packet})};
        {error, _Reason}=Error ->
            Error
    end.


terminate(Dbg
         ,#?STATE{module = Mod
                 ,data = Data
                 ,name = Name
                 ,metadata = #?ETCP_MD{socket = Sock
                                      ,transporter = TrMod
                                      ,options = Opts
                                      ,last_callback = Func
                                      ,last_message = Msg}=TMD}
         ,Rsn) ->
    Rsn2 =
        try Mod:terminate(Rsn, Data, TMD) of
            {error, Rsn3} ->
                Rsn3;
            _ ->
                Rsn
        catch
            _:Rsn3 ->
                [Rsn, {error, {callback_crash, [{reason, Rsn3}
                                               ,{stacktrace, erlang:get_stacktrace()}
                                               ,{module, Mod}
                                               ,{function, terminate}
                                               ,{arguments, [Rsn, Data, TMD]}]}}]
        end,
    ShouldLog =
        try etcp_utils:get_value(error_logger, Opts, fun erlang:is_boolean/1, ?DEF_TERMINATE_LOG) of
            Val ->
                Val
        catch
            _:_ ->
                true
        end,
    if
        ShouldLog ->
            error_logger:format("** ETCP connection ~p terminating \n"
                                "** Reason for termination == ~p\n"
                                "** Last callback == ~p\n"
                                "** Last message == ~p\n"
                                "** State == ~p\n"
                               ,[Name, Rsn2, Func, Msg, Data]);
        true ->
            ok
    end,
    _ = etcp_transporter:close(TrMod, Sock, Opts),
    sys:print_log(Dbg),
    erlang:exit(Rsn2);
terminate(Dbg, _State, Rsn) ->
    sys:print_log(Dbg),
    erlang:exit(Rsn).


connection_debug(_Name, [], _Event) ->
    [];
connection_debug(Name, Dbg, Event) ->
    sys:handle_debug(Dbg, fun connection_handle_debug/3, Name, Event).


connection_handle_debug(IODev, {socket_in, Packet}, Name) ->
    io:format(IODev, "*DBG* ETCP connection ~p got packet ~p\n", [Name, Packet]);
connection_handle_debug(IODev, {socket_out, Packet}, Name) ->
    io:format(IODev, "*DBG* ETCP connection ~p sent packet ~p\n", [Name, Packet]);
connection_handle_debug(IODev, {in, Msg}, Name) ->
    io:format(IODev, "*DBG* ETCP connection ~p got message ~p\n", [Name, Msg]);
connection_handle_debug(IODev, {out, {Pid, Tag}, Msg}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP connection ~p sent ~p to ~p with tag ~p\n"
             ,[Name, Msg, Pid, Tag]);
connection_handle_debug(IODev, {out, Pid, Msg}, Name) ->
    io:format(IODev, "*DBG* ETCP connection ~p sent ~p to ~p\n", [Name, Msg, Pid]);
connection_handle_debug(IODev, {?GEN_CALL_TAG, {Pid, Tag}, Req}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP connection ~p got call ~p from ~p with tag ~p\n"
             ,[Name, Req, Pid, Tag]);
connection_handle_debug(IODev, {?GEN_CAST_TAG, Msg}, Name) ->
    io:format(IODev, "*DBG* ETCP connection ~p got cast ~p\n", [Name, Msg]);
connection_handle_debug(IODev, {?GEN_EVENT_TAG, Event}, Name) ->
    io:format(IODev, "*DBG* ETCP connection ~p got event ~p\n", [Name, Event]);
connection_handle_debug(IODev, {?GEN_ETCP_TAG, {Pid, Tag}, {send, Pkt}}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP connection ~p got synchronous request for sending packet ~p from ~p with t"
              "ag ~p\n"
             ,[Name, Pkt, Pid, Tag]);
connection_handle_debug(IODev, {?GEN_ETCP_TAG, {send, Pkt}}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP connection ~p got asynchronous requestfor sending packet ~p\n"
             ,[Name, Pkt]);
connection_handle_debug(IODev, {system, {Pid, _Tag}, Msg}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP connection ~p got system message '~p' from '~p'"
             ,[Name, Msg, Pid]);
connection_handle_debug(IODev, Event, Name) ->
    io:format(IODev, "*DBG* ETCP connection ~p got debug event ~p\n", [Name, Event]).


acceptor_debug(_Name, [], _Event) ->
    [];
acceptor_debug(Name, Dbg, Event) ->
    sys:handle_debug(Dbg, fun acceptor_handle_debug/3, Name, Event).


acceptor_handle_debug(IODev, {accept, Sock}, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p accepted socket ~p\n", [Name, Sock]);

acceptor_handle_debug(IODev, {start_connection, Sock, Pid}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP acceptor ~p gave socket ~p to connection handler ~p\n"
             ,[Name, Sock, Pid]);

acceptor_handle_debug(IODev, {start_connection_error, Sock, Reason}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP acceptor ~p could not start connection handler for socket ~p with reason "
        "~p\n"
             ,[Name, Sock, Reason]);

acceptor_handle_debug(IODev, {start_connection_ignore, Sock}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP acceptor ~p starting new connection handler for socket ~p was ignored\n"
             ,[Name, Sock]);

acceptor_handle_debug(IODev, {out, Pid, Msg}, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p sent message ~p to ~p \n", [Name, Msg, Pid]);

acceptor_handle_debug(IODev, {?GEN_CALL_TAG, {Pid, _Tag}, Msg}, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p got request ~p from ~p\n", [Name, Msg, Pid]);

acceptor_handle_debug(IODev, {?CHANGE_MODE_TAG, Mode}, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p changed mod to ~p\n", [Name, Mode]);

acceptor_handle_debug(IODev, ?FETCH_MODE_TAG, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p got request for getting its mode\n", [Name]);

acceptor_handle_debug(IODev, {in, Msg}, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p got message ~p\n", [Name, Msg]);

acceptor_handle_debug(IODev, {start, TrMod, ListenSock, ConSup, Mode}, Name) ->
    io:format(IODev
             ,"*DBG* ETCP acceptor ~p started for listen socket ~p with options:\n "
        "\ttransporter module: '~p'\n "
        "\tpool pid: ~p\n "
        "\tmode: '~p'\n"
             ,[Name, ListenSock, TrMod, ConSup, Mode]);

acceptor_handle_debug(IODev, accept_timeout, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p got accept timeout", [Name]);

acceptor_handle_debug(IODev, handshake_timeout, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p got handshake timeout", [Name]);

acceptor_handle_debug(IODev, Event, Name) ->
    io:format(IODev, "*DBG* ETCP acceptor ~p got event ~p\n", [Name, Event]).






fix_activity(TrMod, Sock, Opts) ->
    case etcp_transporter:activity(TrMod, Opts) of
        {ok, false}=Ok ->
            Ok;
        {ok, true} ->
            case etcp_transporter:activate(TrMod, Sock, Opts) of
                ok ->
                    {ok, true};
                {error, _}=Err ->
                    Err
            end;
        {error, _}=Err ->
            Err
    end.


filter_mode(?ACCEPT) ->
    true;
filter_mode(?SLEEP) ->
    true;
filter_mode(_) ->
    false.