%%% ------------------------------------------------------------------------------------------------
%%% ETCP is available for use under the following license, commonly known as the 3-clause (or
%%% "modified") BSD license:
%%%
%%% Copyright (c) 2017-2018, Pouriya Jahanbakhsh
%%% (pouriya.jahanbakhsh@gmail.com)
%%% All rights reserved.
%%%
%%% Redistribution and use in source and binary forms, with or without modification, are permitted
%%% provided that the following conditions are met:
%%%
%%% 1. Redistributions of source code must retain the above copyright notice, this list of
%%%    conditions and the following disclaimer.
%%%
%%% 2. Redistributions in binary form must reproduce the above copyright notice, this list of
%%%    conditions and the following disclaimer in the documentation and/or other materials provided
%%%    with the distribution.
%%%
%%% 3. Neither the name of the copyright holder nor the names of its contributors may be used to
%%%    endorse or promote products derived from this software without specific prior written
%%%    permission.
%%%
%%% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
%%% IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
%%% FITNESS FOR A  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
%%% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
%%% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
%%% SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
%%% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
%%% OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
%%% POSSIBILITY OF SUCH DAMAGE.
%%% ------------------------------------------------------------------------------------------------
%% @author  Pouriya Jahanbakhsh <pouriya.jahanbakhsh@gmail.com>
%% @version 17.9.10
%% @hidden
%% -------------------------------------------------------------------------------------------------
-module(etcp_server_connection_sup).
-author("pouriya.jahanbakhsh@gmail.com").
%% -------------------------------------------------------------------------------------------------
%% Exports:

%% API:
-export([start_link/3
        ,add/2
        ,fetch/2]).

%% 'director' export:
-export([init/1
        ,terminate/2]).

%% -------------------------------------------------------------------------------------------------
%% Recrods & Macros & Includes:

-define(DEF_START_OPTS, []).
-define(DEF_TABLE_TYPE, list).

-include("internal/etcp_guard.hrl").

%% -------------------------------------------------------------------------------------------------
%% Internal functions:

-spec
start_link(module(), etcp_types:init_argument(), etcp_types:start_options()) ->
    etcp_types:start_return().
%% @doc
%%      starts and links connection supervisor for keeping other
%%      connection supervisor(s).
%% @end
start_link(Mod, InitArg ,Opts) when erlang:is_atom(Mod) andalso erlang:is_map(Opts) ->
    director:start_link(?MODULE, {Mod, InitArg, Opts}, ?DEF_START_OPTS).


-spec
add(etcp_types:name(), pos_integer()) ->
    etcp_types:start_return().
%% @doc
%%      adds new supervisor to connection supervisor.
%% @end
add(RootSup, Id) when ?is_proc_ref(RootSup) ->
    director:start_child(RootSup, #{id => Id, append => true}).


-spec
fetch(etcp_types:name(), etcp_types:connection_table_type()) ->
    [] | [{reference(), pid()}].
%% @doc
%%      fetch all available connections from supervisors.
%% @end
fetch(RootSup, list) when ?is_proc_ref(RootSup) ->
    {ok, Pids} = director:get_pids(RootSup),
    etcp_utils:concat([etcp_connection_sup:fetch(Pid, list) || {_Id, Pid} <- Pids]);
fetch(RootSup, {ets, Tab}) when ?is_proc_ref(RootSup) ->
    {ok, Pids} = director_table_ets:get_pids(Tab),
    Pids;
fetch(RootSup, {mnesia, Tab}) when ?is_proc_ref(RootSup) ->
    {ok, Pids} = director_table_ets:get_pids(Tab),
    Pids.

%% -------------------------------------------------------------------------------------------------
%% 'director' callback:

%% @hidden
init({Mod, InitArg, Opts}) ->
    TabType = etcp_utils:get_value(connection_table_type
                                  ,Opts
                                  ,?DEF_TABLE_TYPE
                                  ,fun etcp_utils:filter_table_type/1),
    DefChildSpec = #{start => {etcp_connection_sup, start_link_, [Mod, InitArg, Opts]}
                    ,plan => fun(_, _, _, St) -> {stop, St} end
                    ,type => supervisor
                    ,log_validator => fun child_log_validator/4},
    Opts2 = [{log_validator, fun sup_log_validator/4} | TabType],
    {ok, undefined, [], DefChildSpec, Opts2}.


%% @hidden
terminate(_Rsn, _State) ->
    ok.

%% -------------------------------------------------------------------------------------------------
%% Internal functions:


child_log_validator(_Name, info, start, _State) ->
    none;
child_log_validator(_Name, error, normal, _State) ->
    none;
child_log_validator(_Name, error, _, _State) ->
    long.


sup_log_validator(_Name, error, _, _State) ->
    none;
sup_log_validator(_Name, warning, _, _State) ->
    long.