%%% ------------------------------------------------------------------------------------------------
%%% ETCP is available for use under the following license, commonly known as the 3-clause (or
%%% "modified") BSD license:
%%%
%%% Copyright (c) 2017-2018, Pouriya Jahanbakhsh
%%% (pouriya.jahanbakhsh@gmail.com)
%%% All rights reserved.
%%%
%%% Redistribution and use in source and binary forms, with or without modification, are permitted
%%% provided that the following conditions are met:
%%%
%%% 1. Redistributions of source code must retain the above copyright notice, this list of
%%%    conditions and the following disclaimer.
%%%
%%% 2. Redistributions in binary form must reproduce the above copyright notice, this list of
%%%    conditions and the following disclaimer in the documentation and/or other materials provided
%%%    with the distribution.
%%%
%%% 3. Neither the name of the copyright holder nor the names of its contributors may be used to
%%%    endorse or promote products derived from this software without specific prior written
%%%    permission.
%%%
%%% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
%%% IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
%%% FITNESS FOR A  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
%%% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
%%% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
%%% SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
%%% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
%%% OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
%%% POSSIBILITY OF SUCH DAMAGE.
%%% ------------------------------------------------------------------------------------------------
%% @author  Pouriya Jahanbakhsh <pouriya.jahanbakhsh@gmail.com>
%% @version 17.9.10
%% -------------------------------------------------------------------------------------------------
-module(etcp_server_sup).
-author("pouriya.jahanbakhsh@gmail.com").
%% -------------------------------------------------------------------------------------------------
%% Exports:

%% API:
-export([start_link/3
        ,start_link/4
        ,start_link/5

        ,fetch_connections/2
        ,fetch_acceptors/1

        ,sleep_acceptors/1
        ,wakeup_acceptors/1
        ,fetch_acceptor_modes/1

        ,stop/1
        ,stop/2]).

%% 'director' export:
-export([init/1
        ,terminate/2]).

%% -------------------------------------------------------------------------------------------------
%% Record & Macros & Includes:

-define(ACCEPTOR_SUP, <<"acc_sup">>).
-define(CONNECTION_SUP, <<"con_sup">>).
-define(DEF_DIRECTOR_START_OPTS, []).
-define(DEF_START_OPTS, #{}).
-define(DEF_ACCEPTOR_COUNT, 3).
-define(DEF_SOCKET_OPTIONS, []).
-define(DEF_SSL_FLAG, 'false').
-define(DEF_TERMINATE_TIMEOUT, 'infinity').
-define(DEF_TRANSPORT_MODULE, 'etcp_transporter_tcp').
-define(DEF_SERVER_ERROR_LOGGER, 'undefined').

-include("internal/etcp_guard.hrl").

%% -------------------------------------------------------------------------------------------------
%% API functions:

-spec
start_link(module(), etcp_types:init_argument(), etcp_types:port_number()) ->
     etcp_types:start_return().
start_link(Mod, InitArg, Port) when erlang:is_atom(Mod),
                                    erlang:is_integer(Port) ->
    start_link(Mod, InitArg, Port, ?DEF_START_OPTS).


-spec
start_link(etcp_types:register_name() | module()
          ,module() | etcp_types:init_argument()
          , etcp_types:init_argument() | etcp_types:port_number()
          , etcp_types:port_number() | etcp_types:start_options()) ->
    etcp_types:start_return().
start_link(Mod, InitArg, Port, Opts) when erlang:is_atom(Mod),
                                          erlang:is_integer(Port),
                                          erlang:is_map(Opts) ->
    case director:start_link(?MODULE, {Mod, InitArg, Port, Opts}, ?DEF_DIRECTOR_START_OPTS) of
        {ok, Pid} ->
            continue_starting(Opts, Pid);
        {error, _Reason}=Error ->
            Error;
        ignore ->
            ignore
    end;
start_link(Name, Mod, InitArg, Port) when erlang:is_tuple(Name),
                                          erlang:is_atom(Mod),
                                          erlang:is_integer(Port) ->
    case director:start_link(Name
                            ,?MODULE
                            ,{Mod, InitArg, Port, ?DEF_START_OPTS}
                            ,?DEF_DIRECTOR_START_OPTS) of
        {ok, Pid} ->
            continue_starting(?DEF_START_OPTS, Pid);
        {error, _Reason}=Error ->
            Error;
        ignore ->
            ignore
    end.


-spec
start_link(etcp_types:register_name()
          ,module()
          , etcp_types:init_argument()
          , etcp_types:port_number()
          , etcp_types:start_options()) ->
    etcp_types:start_return().
start_link(Name, Mod, InitArg, Port, Opts) when erlang:is_tuple(Name),
                                                erlang:is_atom(Mod),
                                                erlang:is_integer(Port),
                                                erlang:is_map(Opts) ->
    case director:start_link(Name, ?MODULE, {Mod, InitArg, Port, Opts}, ?DEF_DIRECTOR_START_OPTS) of
        {ok, Pid} ->
            continue_starting(Opts, Pid);
        {error, _Reason}=Error ->
            Error;
        ignore ->
            ignore
    end.


-spec
fetch_acceptors(etcp_types:name()) ->
    [{reference(), pid()}].
fetch_acceptors(Server) when ?is_proc_ref(Server) ->
    {ok, Pid} = director:get_pid(Server, ?ACCEPTOR_SUP),
    etcp_acceptor_root_sup:fetch(Pid).


-spec
fetch_connections(etcp_types:name(), etcp_types:connection_table_type()) ->
    [] | [{reference(), pid()}].
fetch_connections(Server, list) when ?is_proc_ref(Server) ->
    {ok, Pid} = director:get_pid(Server, ?CONNECTION_SUP),
    etcp_server_connection_sup:fetch(Pid, list);
fetch_connections(Server, {ets, Tab}) when ?is_proc_ref(Server) ->
    {ok, Pids} = director_table_ets:get_pids(Tab),
    Pids;
fetch_connections(Server, {mnesia, Tab}) when ?is_proc_ref(Server) ->
    {ok, Pids} = director_table_ets:get_pids(Tab),
    Pids.


-spec
sleep_acceptors(etcp_types:name()) ->
    'ok'.
sleep_acceptors(Server) when ?is_proc_ref(Server) ->
    {ok, Pid} = director:get_pid(Server, ?ACCEPTOR_SUP),
    etcp_acceptor_root_sup:sleep(Pid).


-spec
wakeup_acceptors(etcp_types:name()) ->
    'ok'.
wakeup_acceptors(Server) when ?is_proc_ref(Server) ->
    {ok, Pid} = director:get_pid(Server, ?ACCEPTOR_SUP),
    etcp_acceptor_root_sup:wakeup(Pid).


-spec
fetch_acceptor_modes(etcp_types:name()) ->
    etcp_types:acceptor_mode() | [etcp_types:acceptor_mode()].
fetch_acceptor_modes(Server) when ?is_proc_ref(Server) ->
    {ok, Pid} = director:get_pid(Server, ?ACCEPTOR_SUP),
    etcp_acceptor_root_sup:fetch_mode(Pid).


-spec
stop(etcp_types:name()) ->
    'ok'.
stop(Server) when ?is_proc_ref(Server) ->
    director:stop(Server, normal, ?DEF_TERMINATE_TIMEOUT).


-spec
stop(etcp_types:name(), Reason:: etcp_types:reason()) ->
    'ok'.
stop(Server, Reason) when ?is_proc_ref(Server) ->
    director:stop(Server, Reason, ?DEF_TERMINATE_TIMEOUT).

%% -------------------------------------------------------------------------------------------------
%% 'director' callback:

%% @hidden
init({Mod, InitArg, Port, Opts}=Arg) ->
    TrMod = etcp_utils:get_value(transporter
                                ,Opts
                                ,?DEF_TRANSPORT_MODULE
                                ,fun erlang:is_atom/1),
    ServerErrorLogger = etcp_utils:get_value(server_error_logger
                                            ,Opts
                                            ,?DEF_SERVER_ERROR_LOGGER
                                            ,fun etcp_utils:filter_logger/1),
    case etcp_transporter:listen(TrMod, Port, Opts) of
        {ok, LSock} ->
            OKRet =
                fun(InitArgX) ->
                    State = {ServerErrorLogger, {TrMod, [LSock, Opts]}},
                    Children = [#{id => ?CONNECTION_SUP
                                 ,start => {etcp_server_connection_sup
                                           ,start_link
                                           ,[Mod, InitArgX, Opts]}
                                 ,type => supervisor
                                 ,plan => fun(_, _, _, St) -> {stop, St} end
                                 ,log_validator => fun child_log_validator/4}
                               ,#{id => ?ACCEPTOR_SUP
                                 ,start => {etcp_acceptor_root_sup
                                           ,start_link
                                           ,[Mod, InitArgX, Opts, LSock]}
                                 ,type => supervisor
                                 ,plan => fun(_, _, _, St) -> {stop, St} end
                                 ,log_validator => fun child_log_validator/4}],
                    {ok, State, Children, [{log_validator, fun sup_log_validator/4}]}
                end,
            try Mod:listen_init(InitArg, Port, Opts, LSock) of
                ok ->
                    OKRet(InitArg);
                {ok, InitArg2} ->
                    OKRet(InitArg2);
                ignore ->
                    ignore;
                {stop, _Reason}=Stop ->
                    Stop;
                Other ->
                    {stop, {bad_return, [{returned_value, Other}
                                        ,{module, Mod}
                                        ,{function, listen_init}
                                        ,{arguments, [Arg]}]}}
            catch
                _:Rsn ->
                    {stop, {init_crash, [{reason, Rsn}
                                        ,{module, Mod}
                                        ,{function, listen_init}
                                        ,{arguments, [Arg]}
                                        ,{stacktrace, erlang:get_stacktrace()}]}}
            end;
        {error, Reason} ->
            {stop, Reason}
    end.


%% @hidden
terminate(_Rsn, {_, {Mod, Args}}) ->
    erlang:apply(Mod, close, Args),
    ok.

%% -------------------------------------------------------------------------------------------------
%% Internal functions:


continue_starting(Opts, Pid) ->
    AccCount = etcp_utils:get_value(acceptor_count
                                   ,Opts
                                   ,?DEF_ACCEPTOR_COUNT
                                   ,fun etcp_utils:is_whole_integer/1),
    {ok, ConRootSupPid} = director:get_pid(Pid, ?CONNECTION_SUP),
    ConSups = start_connection_sups(AccCount, ConRootSupPid),
    {ok, AccSupPid} = director:get_pid(Pid, ?ACCEPTOR_SUP),
    ok = start_acceptors(ConSups, AccSupPid),
    {ok, Pid}.


start_connection_sups(Count, ConRootSupPid) ->
    start_connection_sups(Count, ConRootSupPid, []).

start_connection_sups(0, _ConRootSupPid, ConSups) ->
    ConSups;
start_connection_sups(Count, ConRootSupPid, ConSups) ->
    {ok, ConSupPid} = etcp_server_connection_sup:add(ConRootSupPid, Count),
    start_connection_sups(Count-1, ConRootSupPid, [ConSupPid | ConSups]).


start_acceptors(ConSups, AccSup) ->
    lists:foreach(fun(ConSup) -> etcp_acceptor_root_sup:start_acceptor_sup(AccSup, ConSup) end, ConSups).


child_log_validator(_Id, info, start, _State) ->
    none;
child_log_validator(_Id, error, normal, _State) ->
    none;
child_log_validator(_Id, error, _, _State) ->
    long.


sup_log_validator(_Name, error, _Rsn, {undefined, _LSock}) ->
    short;
sup_log_validator(Name, error, Rsn, {Logger, _LSock}) ->
    try
        Logger(Name, Rsn),
        none
    catch
        _:Rsn2 ->
            error_logger:format("ETCP server ~p: Error logger crashed with reason ~p and stacktrace"
                                " ~p\n"
                               ,[Name, Rsn2, erlang:get_stacktrace()]),
            long
    end;
sup_log_validator(_Name, _State, warning, _) ->
    short.