%%% ------------------------------------------------------------------------------------------------
%%% ETCP is available for use under the following license, commonly known as the 3-clause (or
%%% "modified") BSD license:
%%%
%%% Copyright (c) 2017-2018, Pouriya Jahanbakhsh
%%% (pouriya.jahanbakhsh@gmail.com)
%%% All rights reserved.
%%%
%%% Redistribution and use in source and binary forms, with or without modification, are permitted
%%% provided that the following conditions are met:
%%%
%%% 1. Redistributions of source code must retain the above copyright notice, this list of
%%%    conditions and the following disclaimer.
%%%
%%% 2. Redistributions in binary form must reproduce the above copyright notice, this list of
%%%    conditions and the following disclaimer in the documentation and/or other materials provided
%%%    with the distribution.
%%%
%%% 3. Neither the name of the copyright holder nor the names of its contributors may be used to
%%%    endorse or promote products derived from this software without specific prior written
%%%    permission.
%%%
%%% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
%%% IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
%%% FITNESS FOR A  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
%%% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
%%% CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
%%% SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
%%% THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
%%% OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
%%% POSSIBILITY OF SUCH DAMAGE.
%%% ------------------------------------------------------------------------------------------------
%% @author  Pouriya Jahanbakhsh <pouriya.jahanbakhsh@gmail.com>
%% @version 17.9.10
%% @hidden
%% -------------------------------------------------------------------------------------------------
-module(etcp_transporter_tcp).
-author("pouriya.jahanbakhsh@gmail.com").
-behaviour(etcp_transporter).
%% -------------------------------------------------------------------------------------------------
%% Exports:

%% API:
-export([listen/2
        ,accept/2
        ,connect/3
        ,send/3
        ,recv/4
        ,check_message/3
        ,close/2
        ,shutdown/3
        ,controlling_process/3
        ,activity/2
        ,set_options/3
        ,peername/2
        ,get_options/3
        ,receive_message/3
        ,receive_messages/3
        ,activate/2
        ,deactivate/2
        ,activity/1
        ,format_error/1]).

%% -------------------------------------------------------------------------------------------------
%% Records & Macros & Includes:

-define(DEF_ACCEPTOR_ACCEPT_TIMEOUT, 3000).
-define(DEF_SOCKET_OPTIONS, []).
-define(DEF_CONNECT_TIMEOUT, 3000).

%% -------------------------------------------------------------------------------------------------
%% API:

-spec
listen(etcp_types:port_number(), etcp_types:start_options()) ->
    {'ok', etcp_types:socket()} | etcp_types:transporter_callback_error().
listen(Port, Opts) ->
    case get_socket_options(Opts) of
        {ok, SockOpts} ->
            case gen_tcp:listen(Port, [{active, false} | SockOpts] ++ [{active, false}]) of
                {ok, _}=Ok ->
                    Ok;
                {error, Rsn} ->
                    {error, [{reason, Rsn}]}
            end;
        {error, _}=Err ->
            Err
    end.


-spec
accept(etcp_types:socket(), etcp_types:start_options()) ->
    {ok, etcp_types:socket()} | etcp_types:transporter_callback_error().
accept(LSock, Opts) ->
    case gen_tcp:accept(LSock, get_accept_timeout(Opts)) of
        {ok, _}=Ok ->
            Ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
connect(etcp_types:host(), etcp_types:port_number(), etcp_types:start_options()) ->
    {'ok', etcp_types:socket()} | etcp_types:transporter_callback_error().
connect(Host, Port, Opts) ->
    case get_socket_options_and_connect_timeout(Opts) of
        {ok, {SockOpts, CTimeout}} ->
            case gen_tcp:connect(Host, Port, SockOpts, CTimeout) of
                {ok, _}=Ok ->
                    Ok;
                {error, Rsn} ->
                    {error, [{reason, Rsn}]}
            end;
        {error, _}=Err ->
            Err
    end.


-spec
send(etcp_types:socket(), etcp_types:packet(), etcp_types:start_options()) ->
    'ok' | etcp_types:transporter_callback_error().
send(Sock, Packet, _Opts) ->
    case gen_tcp:send(Sock, Packet) of
        ok ->
            ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
recv(etcp_types:socket(), etcp_types:length(), timeout(), etcp_types:start_options()) ->
    {'ok', etcp_types:packet()} | etcp_types:transporter_callback_error().
recv(Sock, Len, Timeout, _Opts) ->
    case gen_tcp:recv(Sock, Len, Timeout) of
        {ok, _}=Ok ->
            Ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
check_message(any(), etcp_types:socket(), etcp_types:start_options()) ->
    {'ok', etcp_types:packet()} | etcp_types:transporter_callback_error() | 'unknown'.
check_message({tcp, Sock, Packet}, Sock, _Opts) ->
    {ok, Packet};
check_message({tcp_closed, Sock}, Sock, _Opts) ->
    {error, [{reason, closed}]};
check_message({tcp_error, Sock, Rsn}, Sock, _Opts) ->
    {error, [{reason, Rsn}]};
check_message(_Other, _Sock, _Opts) ->
    unknown.


-spec
close(etcp_types:socket(), etcp_types:start_options()) ->
    'ok' | etcp_types:transporter_callback_error().
close(Sock, _Opts) ->
    gen_tcp:close(Sock),
    ok.


-spec
shutdown(etcp_types:socket(), etcp_types:shutdown_type(), etcp_types:start_options()) ->
    'ok' | etcp_types:transporter_callback_error().
shutdown(Sock, Type, _Opts) ->
    case gen_tcp:shutdown(Sock, Type) of
        ok ->
            ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
controlling_process(etcp_types:socket(), pid(), etcp_types:start_options()) ->
    'ok' | etcp_types:transporter_callback_error().
controlling_process(Sock, Pid, _Opts) ->
    case gen_tcp:controlling_process(Sock, Pid) of
        ok ->
            ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
activity(etcp_types:socket(), etcp_types:start_options()) ->
    {ok, boolean()} | etcp_types:transporter_callback_error().
activity(Sock, _Opts) ->
    case inet:getopts(Sock, [active]) of
        {ok, [{_, Bool}]} ->
            {ok, Bool};
        {ok, []} ->
            {error, [{reason, unknown}]};
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
set_options(etcp_types:socket(), list(), etcp_types:start_options()) ->
    'ok' | etcp_types:transporter_callback_error().
set_options(Sock, SockOpts, _Opts) ->
    case inet:setopts(Sock, SockOpts) of
        ok ->
            ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
peername(etcp_types:socket(), etcp_types:start_options()) ->
    {'ok', etcp_types:address()} | etcp_types:transporter_callback_error().
peername(Sock, _Opts) ->
    case inet:peername(Sock) of
        {ok, _}=Ok ->
            Ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
get_options(etcp_types:socket(), list(), etcp_types:start_options()) ->
    {'ok', list()} | etcp_types:transporter_callback_error().
get_options(Sock, OptList, _Opts) ->
    case inet:getopts(Sock, OptList) of
        {ok, _}=Ok ->
            Ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


-spec
receive_message(etcp_types:socket(), timeout(), etcp_types:start_options()) ->
    {'ok', etcp_types:packet()} | etcp_types:transporter_callback_error().
receive_message(Sock, Timeout, _Opts) ->
    receive
        {tcp, Sock, Pkt} ->
            {ok, Pkt};
        {tcp_closed, Sock} ->
            {error, [{reason, closed}]};
        {tcp_error, Sock, Rsn} ->
            {error, [{reason, Rsn}]}
    after Timeout ->
        {error, [{reason, timeout}]}
    end.


-spec
receive_messages(etcp_types:socket(), timeout(), etcp_types:start_options()) ->
    {'ok', [etcp_types:packet()]} | etcp_types:transporter_callback_error().
receive_messages(Sock, Timeout, _Opts) ->
    receive_messages2(Sock, Timeout, []).


activate(Sock, _Opts) ->
    case inet:setopts(Sock, [{active, true}]) of
        ok ->
            ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


deactivate(Sock, _Opts) ->
    case inet:setopts(Sock, [{active, false}]) of
        ok ->
            ok;
        {error, Rsn} ->
            {error, [{reason, Rsn}]}
    end.


activity(#{socket_options := SockOpts}) when erlang:is_list(SockOpts) ->
    case lists:keyfind(active, 1, SockOpts) of
        {_, false} ->
            {ok, false};
        _ ->
            {ok, true}
    end;
activity(_Opts) ->
    {ok, true}.


-spec
format_error(any()) ->
    etcp_types:error_info().
format_error(closed) ->
    "Socket connection closed";
format_error(timeout) ->
    "Got timeout";
format_error(Reason) ->
    inet:format_error(Reason).

%% -------------------------------------------------------------------------------------------------
%% Internal functions:

get_accept_timeout(#{acceptor_accept_timeout := ATimeout}) ->
    ATimeout;
get_accept_timeout(_) ->
    ?DEF_ACCEPTOR_ACCEPT_TIMEOUT.


get_socket_options(#{socket_options := SockOpts}) ->
    case etcp_utils:filter_socket_options(SockOpts) of
        ok ->
            {ok, SockOpts};
        {error, {_, ErrParams}} ->
            {error, ErrParams}
    end;
get_socket_options(_) ->
    {ok, ?DEF_SOCKET_OPTIONS}.


get_socket_options_and_connect_timeout(#{socket_options := SockOpts
                                       ,connect_timeout := ConnectTimeout}) ->
    case etcp_utils:filter_socket_options(SockOpts) of
        ok ->
            {ok, {SockOpts, ConnectTimeout}};
        {error, _Reason}=Error ->
            Error
    end;
get_socket_options_and_connect_timeout(#{socket_options := SockOpts}) ->
    {ok, {SockOpts, ?DEF_CONNECT_TIMEOUT}};
get_socket_options_and_connect_timeout(#{connect_timeout := ConnectTimeout}) ->
    {ok, {?DEF_SOCKET_OPTIONS, ConnectTimeout}};
get_socket_options_and_connect_timeout(_) ->
    {ok, {?DEF_SOCKET_OPTIONS, ?DEF_CONNECT_TIMEOUT}}.


receive_messages2(Sock, Timeout, Msgs) ->
    receive
        {tcp, Sock, Pkt} ->
            receive_messages2(Sock, Timeout, [Pkt|Msgs]);
        {tcp_closed, Sock}=Msg ->
            case Msgs of
                [] ->
                    {error, [{reason, closed}]};
                _ ->
                    erlang:self() ! Msg,
                    {ok, lists:reverse(Msgs)}
            end;
        {tcp_error, Sock, Rsn}=Msg ->
            case Msgs of
                [] ->
                    {error, [{reason, Rsn}]};
                _ ->
                    erlang:self() ! Msg,
                    {ok, lists:reverse(Msgs)}
            end
    after Timeout ->
        case Msgs of
            [] ->
                {error, [{reason, timeout}]};
            _ ->
                {ok, lists:reverse(Msgs)}
        end
    end.